@@ -1,5 +1,60 @@
-use std::{collections::HashMap, str::FromStr};
+//! Rust library implementing a custom text generation/templating system. Genex
+//! is similar to [Tracery](https://tracery.io), but with some extra
+//! functionality around using external data.
+//!
+//! # Usage
+//!
+//! First create a grammar, then generate an expansion or multiple expansions
+//! from it.
+//!
+//! ```rust
+//! use std::collections::HashSet;
+//! use std::str::FromStr;
+//! use maplit::hashmap;
+//! use genex::Grammar;
+//!
+//! let grammar = Grammar::from_str(
+//!     r#"
+//!       top = <identifier> <action>
+//!       identifier = [The #occupation#|#name#]
+//!       action = [ran|ate|sat|slept|examined the #patient#]
+//!     "#,
+//! )
+//! .unwrap();
+//!
+//! let data = hashmap! {
+//!     "name".to_string() => "Sam".to_string(),
+//!     "patient".to_string() => "lizard".to_string(),
+//! };
+//!
+//! // Find the top-scoring expansion (the one that uses the most external
+//! // variables):
+//!
+//! let best_expansion = grammar.generate("top", &data).unwrap().unwrap();
+//!
+//! assert_eq!(
+//!     best_expansion,
+//!     "Sam examined the lizard".to_string()
+//! );
+//!
+//! // Get all possible expansions:
+//!
+//! let all_expansions = grammar.generate_all("top", &data).unwrap();
+//!
+//! assert_eq!(
+//!     HashSet::<_>::from_iter(all_expansions),
+//!     HashSet::<_>::from_iter(vec![
+//!         "Sam ran".to_string(),
+//!         "Sam ate".to_string(),
+//!         "Sam sat".to_string(),
+//!         "Sam slept".to_string(),
+//!         "Sam examined the lizard".to_string(),
+//!     ])
+//! );
+//! ```
+use std::{collections::HashMap, rc::Rc, str::FromStr};
 pub mod error;
+mod modifiers;
 mod parser;
 pub use crate::error::Error;
 use itertools::Itertools;
@@ -25,13 +80,37 @@ impl Expansion {
     }
 }
 
+#[derive(Debug, Clone, PartialEq, Eq, Hash)]
+struct VarRef {
+    var: String,
+    modifier: Option<String>,
+}
+
+impl VarRef {
+    #[allow(dead_code)]
+    fn with_variable(var: &str) -> Self {
+        VarRef {
+            var: var.to_string(),
+            modifier: None,
+        }
+    }
+
+    #[allow(dead_code)]
+    fn with_variable_and_modifier(var: &str, modifier: &str) -> Self {
+        VarRef {
+            var: var.to_string(),
+            modifier: Some(modifier.to_string()),
+        }
+    }
+}
+
 #[derive(Debug, Clone, PartialEq, Eq, Hash)]
 enum Node {
     Sequence(Vec<Node>),
     Optional(Box<Node>),
     Choice(Vec<Node>),
     Text(String),
-    VarRef(String),
+    VarRef(VarRef),
     NonTerminal(String),
 }
 
@@ -42,11 +121,20 @@ impl Node {
                 varrefs: vec![],
                 text: text.clone(),
             }]),
-            Node::VarRef(var) => match data.get(var) {
-                Some(value) => Ok(vec![Expansion {
-                    varrefs: vec![var.clone()],
-                    text: value.clone(),
-                }]),
+            Node::VarRef(var) => match data.get(&var.var) {
+                Some(value) => {
+                    let text = match &var.modifier {
+                        Some(modifier) => match grammar.get_modifier(&modifier) {
+                            Some(modifier) => Ok(modifier(value)),
+                            None => Err(Error::UnknownModifierError(modifier.to_string())),
+                        },
+                        None => Ok(value.clone()),
+                    }?;
+                    Ok(vec![Expansion {
+                        varrefs: vec![var.var.clone()],
+                        text: text.clone(),
+                    }])
+                }
                 None => Ok(vec![]),
             },
             Node::NonTerminal(lhs) => match grammar.rules.get(lhs) {
@@ -97,7 +185,10 @@ impl ToString for Node {
             Node::Sequence(children) => {
                 format!("[{}]", children.iter().map(|n| n.to_string()).join(""))
             }
-            Node::VarRef(var) => format!("#{}#", var),
+            Node::VarRef(var) => match &var.modifier {
+                Some(modifier) => format!("#{}|{}#", var.var, modifier),
+                None => format!("#{}#", var.var),
+            },
             Node::NonTerminal(id) => format!("<{}>", id),
             Node::Optional(ref node) => format!("?:[{}]", node.to_string()),
             Node::Choice(nodes) => {
@@ -107,15 +198,17 @@ impl ToString for Node {
     }
 }
 
-#[derive(Debug, Clone)]
+#[derive(Clone)]
 pub struct Grammar {
     rules: HashMap<String, Node>,
+    modifiers: HashMap<String, Rc<dyn Fn(&str) -> String>>,
 }
 
 impl Grammar {
     pub fn new() -> Grammar {
         Grammar {
             rules: HashMap::new(),
+            modifiers: HashMap::new(),
         }
     }
 
@@ -127,6 +220,10 @@ impl Grammar {
         self.rules.get(name)
     }
 
+    pub(crate) fn get_modifier(&self, modifier: &str) -> Option<&dyn Fn(&str) -> String> {
+        self.modifiers.get(modifier).map(|x| x.as_ref())
+    }
+
     pub fn generate_all(&self, name: &str, data: &HashMap<String, String>) -> Result<Vec<String>> {
         let node = self.get_rule(name).unwrap();
         node.expand(self, data)
@@ -143,7 +240,9 @@ impl Grammar {
 
 impl Default for Grammar {
     fn default() -> Self {
-        Grammar::new()
+        let mut grammar = Grammar::new();
+        grammar.modifiers = modifiers::get_default_modifiers();
+        grammar
     }
 }
 
@@ -187,11 +286,14 @@ mod tests {
     use maplit::hashmap;
 
     fn grammar_and_data() -> (Grammar, HashMap<String, String>) {
-        let mut grammar = Grammar::new();
-        grammar.add_rule("location", Node::VarRef("city".to_string()));
+        let mut grammar = Grammar::default();
+        grammar.add_rule(
+            "location",
+            Node::VarRef(VarRef::with_variable_and_modifier("city", "capitalize")),
+        );
         let data = hashmap! {
             "name".to_string() => "John".to_string(),
-            "city".to_string() => "London".to_string(),
+            "city".to_string() => "london".to_string(),
         };
         (grammar, data)
     }
@@ -213,7 +315,7 @@ mod tests {
     #[test]
     fn test_expand_varref() {
         let (grammar, data) = grammar_and_data();
-        let node = Node::VarRef("name".to_string());
+        let node = Node::VarRef(VarRef::with_variable("name"));
         let expansions = node.expand(&grammar, &data).unwrap();
         assert_eq!(
             expansions,
@@ -282,7 +384,7 @@ mod tests {
     fn test_expand_choice() {
         let (grammar, data) = grammar_and_data();
         let snoopy = Node::Text("Snoopy".to_string());
-        let name = Node::VarRef("name".to_string());
+        let name = Node::VarRef(VarRef::with_variable("name"));
         let linus = Node::Text("Linus".to_string());
         let choice = Node::Choice(vec![snoopy, name, linus]);
         let expansions = choice.expand(&grammar, &data).unwrap();
@@ -307,12 +409,12 @@ mod tests {
 
     #[test]
     fn test_to_string() {
-        let mut grammar = Grammar::new();
+        let mut grammar = Grammar::default();
         grammar.add_rule(
             "top",
             Node::Sequence(vec![
                 Node::Text("hi ".to_string()),
-                Node::VarRef("name".to_string()),
+                Node::VarRef(VarRef::with_variable("name")),
                 Node::Text(" in ".to_string()),
                 Node::NonTerminal("location".to_string()),
             ]),
@@ -321,7 +423,7 @@ mod tests {
             "location",
             Node::Sequence(vec![
                 Node::Text("city of ".to_string()),
-                Node::VarRef("city".to_string()),
+                Node::VarRef(VarRef::with_variable("city")),
             ]),
         );
         assert_eq!(
@@ -337,7 +439,7 @@ mod tests {
     fn test_generate() {
         let grammar = Grammar::from_str(
             r#"
-            top = hi <name>?:[, my dear #gender#,] in <location>
+            top = Hi <name>?:[, my dear #gender#,] in <location>.
             name = #name#
             location = [city of #city#|#city# in #county# county]
             "#,
@@ -345,14 +447,19 @@ mod tests {
         .unwrap();
         let data = hashmap! {
             "name".to_string() => "John".to_string(),
-            "city".to_string() => "London".to_string(),
-            "county".to_string() => "Los Angeles".to_string(),
+            "city".to_string() => "Janesville".to_string(),
+            "county".to_string() => "Rock".to_string(),
         };
         let r = grammar.generate("top", &data).unwrap().unwrap();
-        assert_eq!(r, "hi John in London in Los Angeles county");
+        assert_eq!(r, "Hi John in Janesville in Rock county.");
 
-        let exps = grammar.generate_all("top", &data).unwrap();
-        println!("{:#?}", exps);
-        assert_eq!(true, false)
+        let exps = HashSet::<_>::from_iter(grammar.generate_all("top", &data).unwrap());
+        assert_eq!(
+            exps,
+            HashSet::from_iter(vec![
+                "Hi John in Janesville in Rock county.".to_string(),
+                "Hi John in city of Janesville.".to_string(),
+            ])
+        );
     }
 }
